# -*- coding: utf-8 -*-
"""stream plugin"""

__author__  = "Adrien DELLE CAVE"
__license__ = """
    Copyright (C) 2017-2018  fjord-technologies

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""

import bitmath
import boto3
import botocore
import errno
import gc
import json
import logging
import os
import Queue
import re
import resource
import socket
import select
import threading
import time

from datetime import datetime
from dateutil import tz
from dwho.adapters.redis import DWhoAdapterRedis
from dwho.classes.plugins import DWhoPluginBase, PLUGINS
from dwho.config import DWHO_SHARED
from multiprocessing import cpu_count
from sonicprobe import helpers
from sonicprobe.libs import keystore, threading_tcp_server, workerpool, xbstream
from SocketServer import StreamRequestHandler


LOG = logging.getLogger('xtrargon.plugins.stream')

logging.getLogger('boto3').setLevel(logging.WARNING)
logging.getLogger('botocore').setLevel(logging.WARNING)
logging.getLogger('requests').setLevel(logging.WARNING)
logging.getLogger('xbstream').setLevel(logging.WARNING)

DEFAULT_MAX_MEMORY     = '1G'
DEFAULT_MAX_TASKS      = 100
DEFAULT_MAX_WORKERS    = 'auto'
DEFAULT_MAX_LIFETIME   = 300
DEFAULT_SIMULTANEOUS   = 1
DEFAULT_STORAGE_CLASS  = 'STANDARD'
DEFAULT_TIMEOUT        = 60

IBDATA_FILENAME_MATCH  = re.compile(r'^ibdata[0-9]+(?:\.(?:meta|delta))?(?:\.qp)?(?:\.xbcrypt)?$').match

S3_MIN_CHUNK_SIZE      = (5 * 1024 * 1024)
OBJECT_MAX_LIFETIME    = 86400
FINALIZER_MAX_LIFETIME = 86400
PROCESSING_LIFETIME    = 600
RLIMIT_NOFILE          = 65536
_QFINALIZER            = {}
_IBDATA                = {}
_QFINALIZERLOCK        = threading.Lock()
_SYS                   = keystore.Keystore()


class XtrargonLookupKeys(threading.Thread):
    def __init__(self, server_id, smpool, adapter_redis, fsdcfg, killed = False, daemon = True):
        threading.Thread.__init__(self)
        self.server_id     = server_id
        self.smpool        = smpool
        self.adapter_redis = adapter_redis
        self.fsdcfg        = fsdcfg
        self.search_key    = "%s|*" % self.server_id
        self._killed       = killed
        self.daemon        = daemon
        self.processing    = {}

    def clean(self):
        if not self.processing:
            return

        xkeys = self.processing.keys()
        for key in xkeys:
            current_time = time.time()

            if key in self.processing \
               and current_time - self.processing[key] > PROCESSING_LIFETIME:
                del self.processing[key]

        xkeys = None

    def run(self):
        while not self._killed:
            time.sleep(1)
            self.clean()

            LOG.debug("lookup keys: %r", self.search_key)
            for keys in self.adapter_redis.keys(self.search_key).itervalues():
                for key in keys:
                    current_time = time.time()
                    if key in self.processing \
                       and current_time - self.processing[key] < PROCESSING_LIFETIME:
                        continue

                    data = self.adapter_redis.get_key(key)
                    if not data:
                        continue
                    data = json.loads(data)

                    self.processing[key] = current_time
                    XtrargonStreamPlugin.launch_finalizer(data['uid'],
                                                          self.smpool,
                                                          self.adapter_redis,
                                                          self.server_id,
                                                          self.fsdcfg,
                                                          self._killed)

                    _QFINALIZER[data['uid']].put((data['key'], data))


class XtrargonUploadFinalizer(threading.Thread):
    def __init__(self, uid, smpool, adapter_redis, server_id, fsdcfg, killed, daemon = True):
        threading.Thread.__init__(self)
        self.uid           = uid
        self.smpool        = smpool
        self.adapter_redis = adapter_redis
        self.server_id     = server_id
        self.fsdcfg        = fsdcfg
        self.last_time     = 0
        self.processing    = {}
        self._killed       = killed
        self.daemon        = daemon

    def callback(self, key):
        if key in self.processing:
            del self.processing[key]

    def start(self):
        self.last_time  = time.time()
        return threading.Thread.start(self)

    def run(self):
        while not self._killed:
            status = None
            client = DWHO_SHARED.get("xtrargon://%s" % self.uid, 'state')
            if client:
                status = client.get('status')

            try:
                (key, data) = _QFINALIZER[self.uid].get(True, 0.5)
                self.last_time = time.time()
            except Queue.Empty:
                if not self.processing \
                   and status != 'processing' \
                   and self.last_time > 0 \
                   and time.time() > self.last_time + FINALIZER_MAX_LIFETIME:
                    with _QFINALIZERLOCK:
                        del _QFINALIZER[self.uid]
                        if self.uid in _IBDATA:
                            del _IBDATA[self.uid]
                    return
                continue

            current_time = time.time()

            if key in self.processing \
               and current_time - self.processing[key] < PROCESSING_LIFETIME:
                continue

            self.processing[key] = current_time
            self.smpool.run(XtrargonUploadBuilder(self.uid,
                                                  self.adapter_redis,
                                                  self.server_id,
                                                  self.fsdcfg,
                                                  key,
                                                  data),
                            self.callback)


class XtrargonUploadBuilder(object):
    def __init__(self, uid, adapter_redis, server_id, fsdcfg, key, data = None):
        self.uid            = uid
        self.adapter_redis  = adapter_redis
        self.server_id      = server_id
        self.fsdcfg         = fsdcfg
        self.fsdclient      = None
        self.key            = key
        self.data           = data

    def _object_exists(self, key):
        try:
            self.fsdclient.head_object(
                Bucket = self.fsdcfg['bucket'],
                Key    = key)
        except Exception:
            return False
        else:
            return True

    def _copy_object(self, src):
        self.fsdclient.copy_object(
            Bucket       = self.fsdcfg['bucket'],
            Key          = self.key,
            CopySource   = {'Bucket': self.fsdcfg['bucket'],
                            'Key': src},
            StorageClass = self.fsdcfg['storage_class'])

        self.adapter_redis.del_key("%s|%s" % (self.server_id, self.key))

    def _rename_object(self):
        pkey = "%s.1" % self.key

        self.fsdclient.copy_object(
            Bucket       = self.fsdcfg['bucket'],
            Key          = self.key,
            CopySource   = {'Bucket': self.fsdcfg['bucket'],
                            'Key':    pkey},
            StorageClass = self.fsdcfg['storage_class'])

        self.fsdclient.delete_object(
            Bucket     = self.fsdcfg['bucket'],
            Key        = pkey)

        self.adapter_redis.del_key("%s|%s" % (self.server_id, self.key))

    def _create_part(self):
        if self.data['upload_id']:
            return

        m = self.fsdclient.create_multipart_upload(
                Bucket       = self.fsdcfg['bucket'],
                Key          = self.key,
                StorageClass = self.fsdcfg['storage_class'])
        self.data['upload_id'] = m['UploadId']

    def _get_part_size(self, key):
        if key in self.data['files']:
            return self.data['files'][key]

        try:
            obj = self.fsdclient.head_object(
                      Bucket = self.fsdcfg['bucket'],
                      Key    = key)
        except botocore.exceptions.ClientError, e:
            if e.response['Error']['Code'] == '404':
                return
            else:
                raise
        else:
            return obj['ContentLength']

    def _merge_part(self, pkey, i, npart):
        nkey  = "%s.%s" % (self.key, npart)

        psize = self._get_part_size(pkey)
        if psize is not None:
            self.data['files'][pkey] = psize

        nsize = self._get_part_size(nkey)
        if nsize is not None:
            self.data['files'][nkey] = nsize

        if psize is None or nsize is None:
            return False

        if psize <= S3_MIN_CHUNK_SIZE or nsize > S3_MIN_CHUNK_SIZE:
            return

        tmpkey = "%s.%s-%s" % (self.key, i, npart)

        try:
            m      = self.fsdclient.create_multipart_upload(
                         Bucket       = self.fsdcfg['bucket'],
                         Key          = tmpkey,
                         StorageClass = self.fsdcfg['storage_class'])

            tmpupload_id = m['UploadId']
            mupload      = {'Parts': []}

            ppart = self.fsdclient.upload_part_copy(
                        Bucket     = self.fsdcfg['bucket'],
                        Key        = tmpkey,
                        CopySource = {'Bucket': self.fsdcfg['bucket'],
                                      'Key':    pkey},
                        PartNumber = 1,
                        UploadId   = tmpupload_id)
            mupload['Parts'].append({'PartNumber': 1,
                                     'ETag':       ppart['CopyPartResult']['ETag']})

            npart = self.fsdclient.upload_part_copy(
                        Bucket     = self.fsdcfg['bucket'],
                        Key        = tmpkey,
                        CopySource = {'Bucket': self.fsdcfg['bucket'],
                                      'Key':    nkey},
                        PartNumber = 2,
                        UploadId   = tmpupload_id)
            mupload['Parts'].append({'PartNumber': 2,
                                     'ETag':       npart['CopyPartResult']['ETag']})

            self.fsdclient.complete_multipart_upload(
                        Bucket          = self.fsdcfg['bucket'],
                        Key             = tmpkey,
                        UploadId        = tmpupload_id,
                        MultipartUpload = mupload)
        except Exception, e:
            LOG.debug(e)
            return False

        return tmpkey

    def _copy_part(self):
        nb_merged = 0

        self._create_part()

        for x in xrange(1, self.data['nb_part']):
            (i, x)  = (int(x), str(x))
            npart   = i + 1
            merged  = False
            mkey    = None

            if x in self.data['parts']:
                continue

            pkey    = "%s.%s" % (self.key, x)
            pnumber = (i - nb_merged)

            if npart < self.data['nb_part']:
                mkey = self._merge_part(pkey, i, npart)
                if mkey is False:
                    self.data['status'] = 'uncomplete'
                    break
                elif mkey:
                    merged     = True
                    pkey       = mkey
                    nb_merged += 1

            try:
                part = self.fsdclient.upload_part_copy(
                           Bucket     = self.fsdcfg['bucket'],
                           Key        = self.key,
                           CopySource = {'Bucket': self.fsdcfg['bucket'],
                                         'Key':    pkey},
                           PartNumber = pnumber,
                           UploadId   = self.data['upload_id'])

                self.data['parts'][x] = {'PartNumber': pnumber,
                                         'ETag':       part['CopyPartResult']['ETag']}

                if merged:
                    self.data['parts'][str(npart)] = {'PartNumber': None}
            except botocore.exceptions.ClientError, e:
                if e.response['Error']['Code'] == 'NoSuchKey':
                    self.data['status'] = 'uncomplete'
                else:
                    raise
            finally:
                if mkey:
                    self.fsdclient.delete_object(
                        Bucket = self.fsdcfg['bucket'],
                        Key    = mkey)

    def _complete_part(self):
        mupload = {'Parts': []}
        for part in sorted(self.data['parts'].itervalues(),
                           key = lambda k: k['PartNumber']):
            if part['PartNumber'] is not None:
                mupload['Parts'].append(part)

        try:
            self.fsdclient.complete_multipart_upload(
                Bucket          = self.fsdcfg['bucket'],
                Key             = self.key,
                UploadId        = self.data['upload_id'],
                MultipartUpload = mupload)
        except Exception, e:
            self.data['parts'] = {}
            self.adapter_redis.set_key("%s|%s" % (self.server_id, self.key),
                                       json.dumps(self.data))
            LOG.error("unable to complete upload: %r. (key: %r, data: %r)", e, self.key, self.data)
        else:
            for x in xrange(1, self.data['nb_part']):
                try:
                    self.fsdclient.delete_object(
                        Bucket = self.fsdcfg['bucket'],
                        Key    = "%s.%s" % (self.key, x))
                except Exception:
                    continue
            self.adapter_redis.del_key("%s|%s" % (self.server_id, self.key))
            self.data = {}

    def run(self):
        while True:
            try:
                self.fsdclient = boto3.client('s3', config = self.fsdcfg['client'])
                break
            except KeyError:
                time.sleep(0.2)

        value          = None

        try:
            if not self.data:
                value = self.adapter_redis.get_key(self.key)
                if not value:
                    return self.key
                self.data = json.loads(value)

            LOG.info("finalize object: %r", self.key)

            if self._object_exists(self.key) \
               or self.data['timestamp'] + OBJECT_MAX_LIFETIME < time.mktime(time.gmtime()):
                self.adapter_redis.del_key("%s|%s" % (self.server_id, self.key))
                return self.key

            if self.data['ibdata']:
                if self._object_exists(self.data['ibdata']):
                    self._copy_object(self.data['ibdata'])
                return self.key

            if not self.data['nb_part']:
                return self.key

            self.data['status'] = 'processing'

            if self.data['nb_part'] <= 2:
                self._rename_object()
                return self.key

            self._copy_part()

            if self.data['status'] == 'uncomplete':
                raise LookupError("status uncomplete")

            self._complete_part()
        except Exception, e:
            if self.data:
                self.data['status'] = 'error'
                self.data['error']  = repr(e)
                self.adapter_redis.set_key("%s|%s" % (self.server_id, self.key),
                                           json.dumps(self.data))
                _QFINALIZER[self.uid].put((self.key, self.data.copy()))
            LOG.debug(e)
        finally:
            self.data = {}
            (self.fsdclient, value) = (None, None)
            gc.collect()

        return self.key

    def __call__(self):
        return self.run()


class XtrargonWriteFile(object):
    def __init__(self, uid, adapter_redis, server_id, smfobj, dirname, filename, fullname, key, payload, size):
        self.uid           = uid
        self.adapter_redis = adapter_redis
        self.server_id     = server_id
        self.smfobj        = smfobj
        self.dirname       = dirname
        self.filename      = filename
        self.fullname      = fullname
        self.key           = key
        self.payload       = payload
        self.size          = size

    def run(self):
        clear = False

        try:
            (f, smfile) = (None, None)

            fullpath = os.path.join(self.dirname, self.key)
            helpers.make_dirs(os.path.dirname(fullpath))

            LOG.info("write object to file: %r, size: %r", self.key, self.size)

            with open(fullpath, 'wb') as f:
                f.write(self.payload)
                f.flush()
                os.fsync(f.fileno())

            if f:
                f.close()

            with self.smfobj[self.fullname]['lock']:
                self.smfobj[self.fullname]['files'][self.key]  = self.size
                self.smfobj[self.fullname]['up_part']         += 1
                if self.smfobj[self.fullname]['up_part'] == self.smfobj[self.fullname]['nb_part'] - 1:
                    smfile         = self.smfobj[self.fullname].copy()
                    smfile['lock'] = None
                    self.adapter_redis.set_key("%s|%s" % (self.server_id, self.fullname),
                                               json.dumps(smfile))
                    _QFINALIZER[self.uid].put((self.fullname, smfile))
                    clear = True

            LOG.info("file written: %r, size: %r", self.key, self.size)
        except Exception, e:
            LOG.exception(e)
        finally:
            (self.payload, smfile) = (None, None)
            if clear:
                del self.smfobj[self.fullname]

    def __call__(self):
        self.run()


class XtrargonUploadFile(object):
    def __init__(self, uid, adapter_redis, server_id, smfobj, fsdcfg, fullname, key, payload, size):
        self.uid           = uid
        self.adapter_redis = adapter_redis
        self.server_id     = server_id
        self.smfobj        = smfobj
        self.fsdcfg        = fsdcfg
        self.fullname      = fullname
        self.key           = key
        self.payload       = payload
        self.size          = size

    def run(self):
        while True:
            try:
                fsdclient = boto3.client('s3', config = self.fsdcfg['client'])
                break
            except KeyError:
                time.sleep(0.2)

        clear     = False
        try:
            smfile = None

            LOG.info("put object: %r, size: %r", self.key, self.size)

            fsdclient.put_object(Bucket       = self.fsdcfg['bucket'],
                                 Key          = self.key,
                                 Body         = self.payload,
                                 StorageClass = self.fsdcfg['storage_class'])

            with self.smfobj[self.fullname]['lock']:
                self.smfobj[self.fullname]['files'][self.key]  = self.size
                self.smfobj[self.fullname]['up_part']         += 1
                if self.smfobj[self.fullname]['up_part'] == self.smfobj[self.fullname]['nb_part'] - 1:
                    smfile         = self.smfobj[self.fullname].copy()
                    smfile['lock'] = None
                    self.adapter_redis.set_key("%s|%s" % (self.server_id, self.fullname),
                                               json.dumps(smfile))
                    _QFINALIZER[self.uid].put((self.fullname, smfile))
                    clear = True

            LOG.info("object transferred: %r, size: %r", self.key, self.size)
        except Exception, e:
            LOG.exception(e)
        finally:
            (fsdclient, self.payload, smfile) = (None, None, None)
            if clear:
                del self.smfobj[self.fullname]

    def __call__(self):
        self.run()


class XtrargonStreamTx(object):
    def __init__(self, server, client, smfobj, smflock, xbobj, max_memory):
        self.server     = server
        self.client     = client
        self.smfobj     = smfobj
        self.smflock    = smflock
        self.xbobj      = xbobj
        self.max_memory = max_memory
        self.server_id  = self.server.server_id

    def _mk_realdstpath(self, filepath):
        if not self.server.config.get('paths'):
            return filepath

        xpath = None
        for key in sorted(self.server.config['paths'].keys(), key=len, reverse=True):
            path = os.path.commonprefix([filepath, key])
            if path and path in self.server.config['paths']:
                xpath = path
                break

        if not xpath:
            return filepath

        r   = filepath
        cfg = self.server.config['paths'][path]

        if cfg.get('dst') and filepath.startswith(xpath):
            r = os.path.join(cfg['dst'], filepath[len(xpath):])

        return os.path.join(os.path.sep, self.client['client_id'], r.lstrip(os.path.sep))

    def _with_xb_chunk_eof(self, fullname):
        clear = False

        try:
            with self.smfobj[fullname]['lock']:
                self.smfobj[fullname]['nb_part'] = self.xbobj.chunk_no

                if self.smfobj[fullname]['up_part'] == self.smfobj[fullname]['nb_part'] - 1:
                    smfile         = self.smfobj[fullname].copy()
                    smfile['lock'] = None
                    self.server.adapter_redis.set_key("%s|%s" % (self.server_id, fullname),
                                                      json.dumps(smfile))
                    _QFINALIZER[self.client['uid']].put((fullname, smfile))
                    clear = True
            if clear:
                del self.smfobj[fullname]
        finally:
            smfile = None

    def _with_common_ibdata(self, fullname, key):
        try:
            with self.smfobj[fullname]['lock']:
                self.smfobj[fullname]['files'][key]  = self.xbobj.payload_size
                self.smfobj[fullname]['up_part']    += 1
                if self.smfobj[fullname]['up_part'] == self.smfobj[fullname]['nb_part'] - 1:
                    smfile         = self.smfobj[fullname].copy()
                    smfile['lock'] = None
                    self.server.adapter_redis.set_key("%s|%s" % (self.server_id, fullname),
                                                      json.dumps(smfile))
                    _QFINALIZER[self.client['uid']].put((fullname, smfile))
        finally:
            smfile = None

    def run(self):
        fullname = self._mk_realdstpath(os.path.join(self.client['path'], self.xbobj.name)).lstrip('/')
        key      = "%s.%s" % (fullname, self.xbobj.chunk_no)
        filename = "%s.%s" % (self.xbobj.name, self.xbobj.chunk_no)
        ibdata   = None

        with self.smflock:
            if self.client['uid'] in _IBDATA and IBDATA_FILENAME_MATCH(self.xbobj.name):
                if self.xbobj.name not in _IBDATA[self.client['uid']]:
                    _IBDATA[self.client['uid']][self.xbobj.name] = fullname
                else:
                    ibdata = _IBDATA[self.client['uid']][self.xbobj.name]
                    if ibdata == fullname:
                        ibdata = None

            if fullname not in self.smfobj:
                self.smfobj[fullname] = {'date':      datetime.now().replace(microsecond = 0,
                                                                             tzinfo      = tz.tzlocal()).isoformat(),
                                         'timestamp': time.mktime(time.gmtime()),
                                         'uid':       self.client['uid'],
                                         'key':       fullname,
                                         'status':    'new',
                                         'error':     None,
                                         'upload_id': None,
                                         'parts':     {},
                                         'up_part':   0,
                                         'nb_part':   0,
                                         'lock':      threading.Lock(),
                                         'ibdata':    ibdata,
                                         'files':     {}}

        self.smfobj[fullname]['timestamp'] = time.mktime(time.gmtime())

        try:
            if self.xbobj.chunk_type == xbstream.XB_CHUNK_TYPE_EOF:
                self._with_xb_chunk_eof(fullname)
                return

            if ibdata:
                self._with_common_ibdata(fullname, key)
                return

            if _SYS.expired('ps', DEFAULT_TIMEOUT):
                _SYS.set('ps', 'info', helpers.ps_info(os.getpid()))
                if _SYS.get('ps', 'info')['mem_info'].vms >= self.max_memory:
                    gc.collect()

            if _SYS.get('ps', 'info')['mem_info'].vms < self.max_memory:
                func = XtrargonUploadFile(self.client['uid'],
                                          self.server.adapter_redis,
                                          self.server_id,
                                          self.smfobj,
                                          self.server.fsdcfg,
                                          fullname,
                                          key,
                                          self.xbobj.payload,
                                          self.xbobj.payload_size)
            else:
                func = XtrargonWriteFile(self.client['uid'],
                                         self.server.adapter_redis,
                                         self.server_id,
                                         self.smfobj,
                                         self.server.config['write_file_path'],
                                         filename,
                                         fullname,
                                         key,
                                         self.xbobj.payload,
                                         self.xbobj.payload_size)

            self.server.smpool.run(func)
        except Exception, e:
            LOG.exception(e)
        finally:
            self.xbobj.clean()
            self.xbobj = None


class XtrargonStreamHandler(StreamRequestHandler):
    def setup(self):
        StreamRequestHandler.setup(self)
        self.client     = {'client_id':     None,
                           'path':          None,
                           'uid':           None,
                           'simultaneous':  None,
                           'common_ibdata': False}
        self.smfobj     = {}
        self.smflock    = threading.Lock()

        self.max_memory = bitmath.parse_string_unsafe(
                            self.server.config.get('max_memory') or DEFAULT_MAX_MEMORY).to_Byte().value

        self.callbacks  = xbstream.XBStreamDefaultCallbacks(read = self.rfile.read,
                                                            tx   = self._tx)

    def _tx(self, xbobj):
        self.server.smpool.run(XtrargonStreamTx(self.server,
                                                self.client,
                                                self.smfobj,
                                                self.smflock,
                                                xbobj,
                                                self.max_memory).run)

    def handle(self):
        (xbread, r, data) = (None, "", "")

        try:
            while not self.server.killed():
                data = self.rfile.read(1)
                if data == "":
                    break
                r += data
                if r[-2:] == "\r\n":
                    r = r[:-2]
                    break

            client = json.loads(r)
            if not client or not client.get('uid'):
                LOG.error("missing xtrargon uid")
                self.wfile.write("xtrargon: missing xtrargon uid\r\n")
                return

            if not DWHO_SHARED.exists("xtrargon://%s" % client['uid']):
                LOG.error("invalid xtrargon uid: %r", client['uid'])
                self.wfile.write("xtrargon: invalid xtrargon uid\r\n")
                return

            if client.get('status') != 'processing':
                LOG.error("invalid xtrargon status for uid: %r", client['uid'])
                self.wfile.write("xtrargon: invalid xtrargon status\r\n")
                return

            self.wfile.write("xtrargon: OK\r\n")

            self.client.update(client)

            if not client['path']:
                client['path'] = ''

            if self.client['common_ibdata'] \
               and self.client['uid'] not in _IBDATA:
                _IBDATA[self.client['uid']] = {}

            #if self.client['simultaneous']:
            #    self.server.smpool.set_max_workers(int(self.client['simultaneous']))

            XtrargonStreamPlugin.launch_finalizer(self.client['uid'],
                                                  self.server.smpool,
                                                  self.server.adapter_redis,
                                                  self.server.server_id,
                                                  self.server.fsdcfg,
                                                  self.server._killed)

            xbread          = xbstream.XBStreamRead(self.callbacks, set_payload = True)

            while xbread.is_streaming():
                xbread()
        except Exception, e:
            self.wfile.write("xtrargon: exception %r\r\n" % e)
            LOG.exception(e)
        finally:
            (xbread, r, data) = (None, "", "")

    def finish(self):
        StreamRequestHandler.finish(self)
        gc.collect()


class XtrargonStreamServer(threading.Thread):
    def __init__(self, server_id, config, simultaneous, smpool, adapter_redis, fsdcfg):
        threading.Thread.__init__(self)
        config['max_workers'] = simultaneous

        self.server = threading_tcp_server.KillableThreadingTCPServer(
                        config,
                        (config['listen_addr'],
                         config['listen_port']),
                        XtrargonStreamHandler,
                        name = 'stream')
        self.server.server_id     = server_id
        self.server.config        = config
        self.server.smpool        = smpool
        self.server.adapter_redis = adapter_redis
        self.server.fsdcfg        = fsdcfg

    def run(self):
        LOG.info("stream server will now serve")
        while not self.server.killed():
            try:
                self.server.serve_until_killed()
            except (socket.error, select.error), why:
                if errno.EINTR == why[0]:
                    LOG.debug("interrupted system call")
                elif errno.EBADF == why[0] and self.server.killed():
                    LOG.debug("server close")
                else:
                    raise
        LOG.info("stream server exiting")

    def kill(self):
        return self.server.kill()


class XtrargonStreamPlugin(DWhoPluginBase):
    PLUGIN_NAME = 'stream'

    def safe_init(self):
        fsdbucket           = None
        storage_class       = DEFAULT_STORAGE_CLASS

        self.stream_server  = None
        self.fsdcfg         = {}
        self.smpool         = None
        self._killed        = False

        if 'pool' not in self.plugconf:
            self.plugconf['pool'] = {}

        _SYS.set('ps', 'info', helpers.ps_info(os.getpid()))

        resource.setrlimit(resource.RLIMIT_NOFILE, (RLIMIT_NOFILE, RLIMIT_NOFILE))
        resource.setrlimit(resource.RLIMIT_CORE, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))

        if self.plugconf.get('credentials'):
            cred = helpers.load_yaml_file(self.plugconf['credentials'])
            if not cred:
                raise ValueError("unable to read credentials")

            for k, v in cred['s3'].iteritems():
                os.environ[k.upper()] = v

        if os.environ.get('AWS_DEFAULT_BUCKET'):
            fsdbucket = os.environ['AWS_DEFAULT_BUCKET']
        elif self.plugconf.get('bucket'):
            fsdbucket = self.plugconf['bucket']

        if not fsdbucket:
            raise ValueError("missing default bucket setting")

        if os.environ.get('AWS_STORAGE_CLASS'):
            storage_class = os.environ['AWS_STORAGE_CLASS']
        elif self.plugconf.get('storage_class'):
            storage_class = self.plugconf['storage_class']

        config    = self.plugconf
        cfgclient = None

        if config and 'client' in config:
            cfgclient = botocore.client.Config(**config['client'])

        self.fsdcfg         = {'bucket':        fsdbucket,
                               'client':        cfgclient,
                               'storage_class': storage_class}
        self.adapter_redis  = DWhoAdapterRedis(self.config, prefix = 'file_part')

    @staticmethod
    def launch_finalizer(uid, smpool, adapter_redis, server_id, fsdcfg, killed = False):
        if uid in _QFINALIZER:
            return

        with _QFINALIZERLOCK:
            LOG.info("launching upload finalizer. (uid: %r)", uid)
            _QFINALIZER[uid] = Queue.Queue()
            XtrargonUploadFinalizer(uid,
                                    smpool,
                                    adapter_redis,
                                    server_id,
                                    fsdcfg,
                                    killed).start()

    def _set_smpool(self):
        max_workers     = self.plugconf['pool'].get('max_workers', DEFAULT_MAX_WORKERS) or 1
        if max_workers == 'auto':
            max_workers = cpu_count()
        elif max_workers < 1:
            max_workers = 1

        worker_lifetime = int(self.plugconf['pool'].get('worker_lifetime', DEFAULT_MAX_LIFETIME))
        max_tasks       = int(self.plugconf['pool'].get('max_tasks', DEFAULT_MAX_TASKS))
        self.smpool     = workerpool.WorkerPool(max_workers = int(max_workers),
                                                life_time   = worker_lifetime,
                                                max_tasks   = max_tasks,
                                                name        = 'smpool')

    def _start_stream_server(self, simultaneous):
        self.stream_server  = XtrargonStreamServer(self.config['general']['server_id'],
                                                   self.plugconf,
                                                   simultaneous,
                                                   self.smpool,
                                                   self.adapter_redis,
                                                   self.fsdcfg)
        self.stream_server.daemon = True
        self.stream_server.start()

    def at_start(self):
        simultaneous                  = int(self.plugconf.get('simultaneous', DEFAULT_SIMULTANEOUS) or 1)
        self.plugconf['simultaneous'] = simultaneous
        # streaming workers: xtrabackup + mysqldump
        simultaneous += 1

        self._set_smpool()

        XtrargonLookupKeys(self.config['general']['server_id'],
                           self.smpool,
                           self.adapter_redis,
                           self.fsdcfg,
                           self._killed).start()

        self._start_stream_server(simultaneous)

    def at_stop(self):
        self._killed = True

        if self.stream_server:
            self.stream_server.kill()

        for queue in _QFINALIZER.itervalues():
            with queue.mutex:
                queue.queue.clear()

        if self.smpool:
            self.smpool.killall(1)


if __name__ != "__main__":
    def _start():
        PLUGINS.register(XtrargonStreamPlugin())
    _start()
